#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Launch gaitutils menu.

@author: Jussi (jnu@iki.fi)
"""
from __future__ import print_function


from gaitutils.gui import _gaitmenu

_gaitmenu.main()
